"""Tests for SVD functionality."""

from matfree import decomp, eig, test_util
from matfree.backend import linalg, np, testing


@testing.parametrize("nrows", [10])
@testing.parametrize("ncols", [3])
def test_equal_to_linalg_svd(nrows, ncols):
    """The output of full-depth SVD should be equal (*) to linalg.svd().

    (*) Note: The singular values should be identical,
    and the orthogonal matrices should be orthogonal. They are not unique.
    """
    d = np.arange(1.0, 1.0 + min(nrows, ncols))
    A = test_util.asymmetric_matrix_from_singular_values(d, nrows=nrows, ncols=ncols)
    v0 = np.ones((ncols,))
    num_matvecs = min(nrows, ncols)

    bidiag = decomp.bidiag(num_matvecs)
    svd = eig.svd_partial(bidiag)
    Ut, S, Vt = svd(lambda v, p: p @ v, v0, A)

    U_, S_, Vt_ = linalg.svd(A, full_matrices=False)

    tols_decomp = {"atol": 1e-5, "rtol": 1e-5}
    assert np.allclose(S, S_, **tols_decomp)
    assert np.allclose(Ut.T @ Ut, U_ @ U_.T, **tols_decomp)
    assert np.allclose(Vt @ Vt.T, Vt_ @ Vt_.T, **tols_decomp)


@testing.parametrize("nrows", [10])
@testing.parametrize("ncols", [3])
@testing.parametrize("num_matvecs", [0, 2, 3])
def test_shapes_as_expected_vectors(nrows, ncols, num_matvecs):
    d = np.arange(1.0, 1.0 + min(nrows, ncols))
    A = test_util.asymmetric_matrix_from_singular_values(d, nrows=nrows, ncols=ncols)
    v0 = np.ones((ncols,))

    bidiag = decomp.bidiag(num_matvecs)
    svd = eig.svd_partial(bidiag)
    Ut, S, Vt = svd(lambda v, p: p @ v, v0, A)

    assert Ut.shape == (num_matvecs, nrows)
    assert S.shape == (num_matvecs,)
    assert Vt.shape == (num_matvecs, ncols)


@testing.parametrize("nrows", [10])
@testing.parametrize("num_matvecs", [0, 2, 3])
def test_shapes_as_expected_lists_tuples(nrows, num_matvecs):
    K = np.arange(1.0, 10.0).reshape((3, 3))
    v0 = np.ones((nrows, nrows))  # tensor-valued input

    # Map Pytrees to Pytrees
    def Av(v: tuple, stencil) -> list:
        (x,) = v
        return [np.convolve2d(stencil, x)]

    bidiag = decomp.bidiag(num_matvecs)
    svd = eig.svd_partial(bidiag)
    [Ut], S, (Vt,) = svd(Av, (v0,), K)

    # Ut inherits pytree-shape from the outputs (list),
    # and Vt inherits pytree-shape from the inputs (tuple)
    [u0] = Av((v0,), K)
    assert Ut.shape == (num_matvecs, *u0.shape)
    assert S.shape == (num_matvecs,)
    assert Vt.shape == (num_matvecs, *v0.shape)



"""Generate API docs from the source."""

import pathlib


def path_as_module(p: pathlib.Path) -> str:
    """Turn a file-path to a python-module-like string."""
    p_as_string = str(p)
    return p_as_string.replace(".py", "").replace("/", ".")


def path_as_markdown_file(p: pathlib.Path) -> str:
    """Turn a file-path to a markdown-file-name."""
    return p.relative_to(path_source).stem


if __name__ == "__main__":
    # Read the arguments
    path_source = pathlib.Path("matfree")
    path_skip = "backend/*"
    path_target = pathlib.Path("docs/API_documentation/")

    # Make the target directory unless it exists
    path_target.mkdir(parents=True, exist_ok=True)

    # Loop recursively through the source
    for path in path_source.rglob("*.py"):
        # Skip "private" modules and
        # skip required files. (E.g., backend/*)
        if path.name[0] != "_" and not path.match(path_skip):
            # Construct the contents
            p_as_module = path_as_module(path)
            p_as_markdown = path_as_markdown_file(path)

            # Open the markdown file and write content
            with open(f"{path_target}/{p_as_markdown}.md", "w") as file:
                content = f"# {p_as_module} \n\n:::{p_as_module}"
                file.write(content)



"""Transform py:light scripts into Jupyter notebooks."""

import contextlib
import os
from re import sub


def py_to_py_light(*, source, target):
    """Transform a Python script into py:light (as expected by JupyText).

    https://jupytext.readthedocs.io/en/latest/formats-scripts.html#the-light-format
    """
    for dir_src, _children, files_src in os.walk(source):
        # Source-directory to target directory
        dir_trg = replace_prefix(dir_src, source=source, target=target)

        # Create directory if necessary
        mkdir_unless_exists(dir_trg)

        # Transform each python file in directory
        for i, filename_with_py in enumerate(sorted(files_src)):
            if filename_with_py[0] != "_" and filename_with_py[-3:] == ".py":
                # Split docstring (to be manipulated) from rest of source
                header, rest = split_docstring_from_content(
                    f"{dir_src}/{filename_with_py}"
                )

                # If first line ends on '.' (which it should), remove ".".
                # We implement this by removing all "."'s from the string.
                # More than one period is not supposed
                # to be in the first-line docstring.
                header[0] = header[0].replace(".", "")

                # Turn header into markdown
                markdown_title = [f"## {header[0]}"]
                markdown_rest = [f"# {h}" for h in header[1:]]
                markdown = [*markdown_title, *markdown_rest]

                # Save results
                title = filename(header[0], remove=["(", ")", "[", "]"])
                save_lines_as_file(
                    [*markdown, "\n", *rest], target=f"{dir_trg}/{i + 1}_{title}.py"
                )


def split_docstring_from_content(file):
    """Split a file into two parts: docstring and content."""
    with open(file) as original:
        line_first, *lines_other = original.readlines()

        if line_first[:2] == 'r"' or line_first[:2] == 'f"':
            line_first = line_first[1:]

        # Remove leading '"""'
        line_first = line_first[3:]

        # Case 1: Single-line docstring
        if line_first[-4:] == '"""\n':
            return [line_first[:-4]], lines_other

        # Case 2: Multi-line docstring
        idx = lines_other.index('"""\n')
        return [line_first, *lines_other[:idx]], lines_other[idx + 1 :]


def save_lines_as_file(data, /, *, target):
    """Save a list of strings as a file."""
    with open(target, "w") as modified:
        x = data[0]
        for remaining in data[1:]:
            x += f"{remaining}"
        modified.write(x)


def replace_prefix(f: str, *, source: str, target: str) -> str:
    """Replace the source-prefix with a target prefix."""
    f_without_source = string_without_prefix(f, prefix=source)
    return f"{target}{f_without_source}"


def string_without_prefix(f: str, /, *, prefix: str) -> str:
    """Remove a prefix from a string."""
    return f[len(prefix) :]


def mkdir_unless_exists(f: str, /) -> None:
    """Create a directory."""
    with contextlib.suppress(FileExistsError):
        os.mkdir(f)


def filename(s, remove):
    """Turn a string into a file-name.

    Removes a list of sub-strings and returns a snake_case version.
    """
    for r in remove:
        s = s.replace(r, "")
    return snake_case(s)


def snake_case(s):
    """Convert a string to snake case.

    Source:
    https://www.w3resource.com/python-exercises/string/python-data-type-string-exercise-97.php
    """
    # Replace hyphens with spaces,
    # then apply regular expression substitutions for title case conversion
    # and add an underscore between words,
    # finally convert the result to lowercase
    return "_".join(
        sub(
            "([A-Z][a-z]+)", r" \1", sub("([A-Z]+)", r" \1", s.replace("-", " "))
        ).split()
    ).lower()


if __name__ == "__main__":
    mkdir_unless_exists("docs/Tutorials/")
    py_to_py_light(source="tutorials/", target="docs/Tutorials/")



"""Tests for Hessenberg factorisations (-> Arnoldi)."""

from matfree import decomp, test_util
from matfree.backend import linalg, np, prng, testing


@testing.parametrize("nrows", [10])
@testing.parametrize("num_matvecs", [0, 5, 9])
@testing.parametrize("reortho", ["none", "full"])
@testing.parametrize("dtype", [float])
def test_decomposition_is_satisfied(nrows, num_matvecs, reortho, dtype):
    # Create a well-conditioned test-matrix
    A = prng.normal(prng.prng_key(1), shape=(nrows, nrows), dtype=dtype)
    v = prng.normal(prng.prng_key(2), shape=(nrows,), dtype=dtype)

    # Decompose
    algorithm = decomp.hessenberg(num_matvecs, reortho=reortho)
    Q, H, r, c = algorithm(lambda s, p: p @ s, v, A)

    # Assert shapes
    assert Q.shape == (nrows, num_matvecs)
    assert H.shape == (num_matvecs, num_matvecs)
    assert r.shape == (nrows,)
    assert c.shape == ()

    # Test the decompositions
    e0, ek = np.eye(num_matvecs)[[0, -1], :]
    test_util.assert_allclose(A @ Q - Q @ H - linalg.outer(r, ek), 0.0)
    test_util.assert_allclose(Q.T.conj() @ Q - np.eye(num_matvecs), 0.0)

    if num_matvecs > 0:
        test_util.assert_allclose(Q @ e0, c * v)


@testing.parametrize("nrows", [10])
@testing.parametrize("num_matvecs", [5, 9])
@testing.parametrize("reortho", ["full"])
def test_reorthogonalisation_improves_the_estimate(nrows, num_matvecs, reortho):
    # Create an ill-conditioned test-matrix (that requires reortho=True)
    A = linalg.hilbert(nrows)
    v = prng.normal(prng.prng_key(2), shape=(nrows,))

    # Decompose
    algorithm = decomp.hessenberg(num_matvecs, reortho=reortho)
    Q, H, r, c = algorithm(lambda s, p: p @ s, v, A)

    # Assert shapes
    assert Q.shape == (nrows, num_matvecs)
    assert H.shape == (num_matvecs, num_matvecs)
    assert r.shape == (nrows,)
    assert c.shape == ()

    # Test the decompositions
    e0, ek = np.eye(num_matvecs)[[0, -1], :]
    test_util.assert_allclose(A @ Q - Q @ H - linalg.outer(r, ek), 0.0)
    test_util.assert_allclose(Q.T @ Q - np.eye(num_matvecs), 0.0)
    test_util.assert_allclose(Q @ e0, c * v)


@testing.parametrize("reortho_wrong", [True, "full_with_sparsity", "None"])
def test_raises_error_for_wrong_reorthogonalisation_flag(reortho_wrong):
    with testing.raises(TypeError, match="Unexpected input"):
        _ = decomp.hessenberg(1, reortho=reortho_wrong)



"""Compute matrix functions without materializing large matrices.

Sometimes, we need to compute matrix exponentials, log-determinants,
or similar functions of matrices, but our matrices are too big to
use functions from
[scipy.linalg](https://docs.scipy.org/doc/scipy/reference/linalg.html)
or
[jax.scipy.linalg](https://jax.readthedocs.io/en/latest/jax.scipy.html#module-jax.scipy.linalg).
However, matrix-free linear algebra scales to even the largest of matrices.
Here is how to use Matfree to compute functions of large matrices.
"""

import functools

import jax

from matfree import decomp, funm

n = 7  # imagine n = 10^5 or larger

key = jax.random.PRNGKey(1)
key, subkey = jax.random.split(key, num=2)
large_matrix = jax.random.normal(subkey, shape=(n, n))


# The expected value is computed with jax.scipy.linalg.

key, subkey = jax.random.split(key, num=2)
vector = jax.random.normal(subkey, shape=(n,))
expected = jax.scipy.linalg.expm(large_matrix) @ vector
print(expected)


# Instead of using jax.scipy.linalg, we can use matrix-vector products
# in combination with the Arnoldi iteration to approximate the
# matrix-function-vector product.


def large_matvec(v):
    """Evaluate a matrix-vector product."""
    return large_matrix @ v


num_matvecs = 5
arnoldi = decomp.hessenberg(num_matvecs, reortho="full")
dense_funm = funm.dense_funm_pade_exp()
matfun_vec = funm.funm_arnoldi(dense_funm, arnoldi)
received = matfun_vec(large_matvec, vector)
print(received)


# The matrix-function vector product can be combined with all usual
# JAX transformations. For example, after fixing the matvec-function
# as the first argument, we can vectorize the matrix function with jax.vmap
# and compile it with jax.jit.

matfun_vec = functools.partial(matfun_vec, large_matvec)
key, subkey = jax.random.split(key, num=2)
vector_batch = jax.random.normal(subkey, shape=(5, n))  # a batch of 5 vectors
received = jax.jit(jax.vmap(matfun_vec))(vector_batch)
print(received.shape)


# Talking about function transformations: we can also
# reverse-mode-differentiate the matrix functions efficiently.

jac = jax.jacrev(matfun_vec)(vector)
print(jac)

# Under the hood, reverse-mode derivatives of Arnoldi- and Lanczos-based
# matrix functions use the fast algorithm for gradients of the
# Lanczos and Arnoldi iterations from
# [this paper](https://arxiv.org/abs/2405.17277).
# Please consider citing it if you use reverse-mode derivatives
# functions of matrices
# (a BibTex is [here](https://pnkraemer.github.io/matfree/)).



"""Numerical linear algebra."""

import jax
import jax.numpy as jnp


def vector_norm(x, /):
    return jnp.linalg.norm(x)


def matrix_norm(x, /, which):
    return jnp.linalg.norm(x, ord=which)


def qr_reduced(x, /):
    return jnp.linalg.qr(x, mode="reduced")


def eigh(x, /):
    return jnp.linalg.eigh(x)


def eig(x, /):
    vals, vecs = jnp.linalg.eig(x)

    # Order the values and vectors by magnitude
    # (jax.numpy.linalg.eig does not do this)
    ordered = jnp.argsort(vals)[::-1]
    vals = vals[ordered]
    vecs = vecs[:, ordered]
    return vals, vecs


def cholesky(x, /):
    return jnp.linalg.cholesky(x)


def cho_factor(matrix, /):
    return jax.scipy.linalg.cho_factor(matrix)


def cho_solve(factor, b, /):
    return jax.scipy.linalg.cho_solve(factor, b)


def slogdet(x, /):
    return jnp.linalg.slogdet(x)


def inner(x1, x2, /):
    # todo: distinguish vecdot, vdot, dot, and matmul?
    return jnp.inner(x1, x2)


def outer(a, b, /):
    return jnp.outer(a, b)


def hilbert(n, /):
    return jax.scipy.linalg.hilbert(n)


def diagonal(x, /, offset=0):
    """Extract the diagonal of a matrix."""
    return jnp.diag(x, offset)


def diagonal_matrix(x, /, offset=0):  # not part of array API
    """Construct a diagonal matrix."""
    return jnp.diag(x, offset)


def trace(x, /):
    return jnp.trace(x)


def svd(A, /, *, full_matrices=True):
    return jnp.linalg.svd(A, full_matrices=full_matrices)


def pinv(A, /):
    return jnp.linalg.pinv(A)


def solve(A, b, /):
    return jnp.linalg.solve(A, b)


def cg(Av, b, /):
    return jax.scipy.sparse.linalg.cg(Av, b)


def funm_schur(A, f, /):
    return jax.scipy.linalg.funm(A, f)


def funm_pade_exp(A, /):
    return jax.scipy.linalg.expm(A)


def lstsq(A, b):
    return jnp.linalg.lstsq(A, b)



"""Estimate log-determinants of PyTree-valued functions.

Can we compute log-determinants if the matrix-vector
products are pytree-valued?
Yes, we can. Matfree natively supports PyTrees.
"""

import jax
import jax.numpy as jnp

from matfree import decomp, funm, stochtrace

# Create a test-problem: a function that maps a pytree (dict) to a pytree (tuple).
# Its (regularised) Gauss--Newton Hessian shall be the matrix-vector product
# whose log-determinant we estimate.


def testfunc(x):
    """Map a dictionary to a tuple with some arbitrary values."""
    return jnp.linalg.norm(x["weights"]), x["bias"]


# Create a test-input

b = jnp.arange(1.0, 40.0)
W = jnp.stack([b + 1.0, b + 2.0])
x0 = {"weights": W, "bias": b}

# Linearise the functions

f0, jvp = jax.linearize(testfunc, x0)
_f0, vjp = jax.vjp(testfunc, x0)
print(jax.tree.map(jnp.shape, f0))
print(jax.tree.map(jnp.shape, jvp(x0)))
print(jax.tree.map(jnp.shape, vjp(f0)))


# Use the same API as if the matrix-vector product were array-valued.
# Matfree flattens all trees internally.


def make_matvec(alpha):
    """Create a matrix-vector product function."""

    def fun(fx, /):
        r"""Matrix-vector product with $J J^\top + \alpha I$."""
        vjp_eval = vjp(fx)
        matvec_eval = jvp(*vjp_eval)
        return jax.tree.map(lambda x, y: x + alpha * y, matvec_eval, fx)

    return fun


matvec = make_matvec(alpha=0.1)
num_matvecs = 3
tridiag_sym = decomp.tridiag_sym(num_matvecs)
integrand = funm.integrand_funm_sym_logdet(tridiag_sym)
sample_fun = stochtrace.sampler_normal(f0, num=10)
estimator = stochtrace.estimator(integrand, sampler=sample_fun)
key = jax.random.PRNGKey(1)
logdet = estimator(matvec, key)
print(logdet)


# For reference: flatten all arguments
# and compute the dense log-determinant:

f0_flat, unravel_func_f = jax.flatten_util.ravel_pytree(f0)


def make_matvec_flat(alpha):
    """Create a flattened matrix-vector-product function."""

    def fun(f_flat):
        """Evaluate a flattened matrix-vector product."""
        f_unravelled = unravel_func_f(f_flat)
        vjp_eval = vjp(f_unravelled)
        matvec_eval = jvp(*vjp_eval)
        f_eval, _unravel_func = jax.flatten_util.ravel_pytree(matvec_eval)
        return f_eval + alpha * f_flat

    return fun


matvec_flat = make_matvec_flat(alpha=0.1)
M = jax.jacfwd(matvec_flat)(f0_flat)
print(jnp.linalg.slogdet(M))



"""Matrix-free bounds on functions of matrices."""

from matfree.backend import linalg, np


def baigolub96_logdet_spd(bound_spectrum, /, nrows, trace, norm_frobenius_squared):
    """Bound the log-determinant of a symmetric, positive definite matrix.

    This function implements Theorem 2 in the paper by Bai and Golub (1996).

    ``bound_spectrum`` is either an upper or a lower bound
    on the spectrum of the matrix.
    If it is an upper bound,
    the function returns an upper bound of the log-determinant.
    If it is a lower bound,
    the function returns a lower bound of the log-determinant.



    ??? note "BibTex for Bai and Golub (1996)"
        ```bibtex
        @article{bai1996bounds,
            title={Bounds for the trace of the inverse and the
            determinant of symmetric positive definite matrices},
            author={Bai, Zhaojun and Golub, Gene H},
            journal={Annals of Numerical Mathematics},
            volume={4},
            pages={29--38},
            year={1996},
            publisher={Citeseer}
        }
        ```
    """
    mu1, mu2 = trace, norm_frobenius_squared
    beta = bound_spectrum
    tbar = (beta * mu1 - mu2) / (beta * nrows - mu1)
    v = np.asarray([np.log(beta), np.log(tbar)])
    w = np.asarray([mu1, mu2])
    A = np.asarray([[beta, tbar], [beta**2, tbar**2]])
    return v @ linalg.solve(A, w)



"""Test the estimation with multiple statistics."""

from matfree import stochtrace
from matfree.backend import func, linalg, np, prng, testing, tree


def test_yields_correct_tree_structure():
    """Assert that mean and standard-deviation are estimated correctly."""

    def fun(x):
        """Create a nonlinear, to-be-differentiated function."""
        fx = np.sin(np.flip(np.cos(x["params"])) + 1.0) * np.sin(x["params"])
        return {"params": fx}

    key = prng.prng_key(seed=2)

    # Linearise function
    x0 = prng.uniform(key, shape=(4,))  # random lin. point
    args_like = {"params": x0}
    _, jvp = func.linearize(fun, args_like)

    # Estimate the matrix function
    problem = stochtrace.integrand_diagonal()
    problem = stochtrace.integrand_wrap_moments(
        problem, moments={"moment_1st": 1, "moment_2nd": 2}
    )
    sampler = stochtrace.sampler_normal(args_like, num=100_000)
    estimate = stochtrace.estimator(problem, sampler=sampler)
    received = estimate(jvp, key)

    irrelevant_value = 1.1
    expected = {
        "params": {"moment_1st": irrelevant_value, "moment_2nd": irrelevant_value}
    }
    assert tree.tree_structure(received) == tree.tree_structure(expected)


@testing.fixture(name="key")
def fixture_key():
    """Fix a pseudo-random number generator."""
    return prng.prng_key(seed=1)


@testing.fixture(name="J_and_jvp")
def fixture_J_and_jvp(key):
    """Create a nonlinear, to-be-differentiated function."""

    def fun(x):
        return np.sin(np.flip(np.cos(x)) + 1.0) * np.sin(x) + 1.0

    # Linearise function
    x0 = prng.uniform(key, shape=(2,))  # random lin. point
    _, jvp = func.linearize(fun, x0)
    J = func.jacfwd(fun)(x0)

    return J, jvp, x0


def test_yields_correct_variance_normal(J_and_jvp, key):
    """Assert that the estimated trace approximates the true trace accurately."""
    # Estimate the trace
    J, jvp, args_like = J_and_jvp
    problem = stochtrace.integrand_trace()
    problem = stochtrace.integrand_wrap_moments(problem, moments=[1, 2])
    sampler = stochtrace.sampler_normal(args_like, num=1_000_000)
    estimate = stochtrace.estimator(problem, sampler=sampler)
    first, second = estimate(jvp, key)

    # Assert the trace is correct
    truth = linalg.trace(J)
    assert np.allclose(first, truth, rtol=1e-2)

    # Assert the variance is correct:
    norm = linalg.matrix_norm(J, which="fro") ** 2
    assert np.allclose(second - first**2, norm * 2, rtol=1e-2)


def test_yields_correct_variance_rademacher(J_and_jvp, key):
    """Assert that the estimated trace approximates the true trace accurately."""
    # Estimate the trace
    J, jvp, args_like = J_and_jvp

    problem = stochtrace.integrand_trace()
    problem = stochtrace.integrand_wrap_moments(problem, moments=[1, 2])
    sampler = stochtrace.sampler_rademacher(args_like, num=500)
    estimate = stochtrace.estimator(problem, sampler=sampler)
    first, second = estimate(jvp, key)

    # Assert the trace is correct
    truth = linalg.trace(J)
    assert np.allclose(first, truth, rtol=1e-2)

    # Assert the variance is correct:
    norm = linalg.matrix_norm(J, which="fro") ** 2
    truth = norm - linalg.trace(J**2)
    assert np.allclose(second - first**2, truth, atol=1e-2, rtol=1e-2)



"""Test matrix-polynomial-vector algorithms via Chebyshev's recursion."""

from matfree import funm, test_util
from matfree.backend import linalg, np, prng


def test_funm_chebyshev(n=12):
    """Test matrix-polynomial-vector algorithms via Chebyshev's recursion."""
    # Create a test-problem: matvec, matrix function,
    # vector, and parameters (a matrix).

    def matvec(x, p):
        return p @ x

    def fun(x):
        return np.sin(x)

    v = prng.normal(prng.prng_key(2), shape=(n,))

    eigvals = np.linspace(-1 + 0.01, 1 - 0.01, num=n)
    matrix = test_util.symmetric_matrix_from_eigenvalues(eigvals)

    # Compute the solution
    eigvals, eigvecs = linalg.eigh(matrix)
    log_matrix = eigvecs @ linalg.diagonal(fun(eigvals)) @ eigvecs.T
    expected = log_matrix @ v

    # Create an implementation of the Chebyshev-algorithm
    num_matvecs = 6
    matfun_vec = funm.funm_chebyshev(fun, num_matvecs, matvec)

    # Compute the matrix-function vector product
    received = matfun_vec(v, matrix)
    assert np.allclose(expected, received, rtol=1e-4)



"""Matrix-free algorithms for least-squares-type problems."""

from matfree.backend import containers, control_flow, func, linalg, np, tree
from matfree.backend.typing import Array, Callable

# todo: make functions compatible with pytree-valued vecmats


_LARGE_VALUE = 1e10
"""A placeholder for np.inf.

In stopping criteria, if we divide something by something that's almost zero,
we don't want the result to be infinite
but only 'large enough to terminate the iteration'.
"""


def lsmr(
    *,
    atol: float = 1e-6,
    btol: float = 1e-6,
    ctol: float = 1e-8,
    maxiter: int = 1_000_000,
    while_loop=control_flow.while_loop,
):
    """Construct an experimental implementation of LSMR.

    Follows the [implementation in SciPy](https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.lsmr.html),
    but uses JAX.
    """
    # todo: stop iteration when NaN or Inf are detected.

    @tree.register_dataclass
    @containers.dataclass
    class State:
        """LSMR state."""

        # Iteration count:
        itn: int
        # Bidiagonalisation variables:
        alpha: float
        u: Array
        v: Array
        # LSMR-specific variables:
        alphabar: float
        rhobar: float
        rho: float
        zeta: float
        sbar: float
        cbar: float
        zetabar: float
        hbar: Array
        h: Array
        x: Array
        # Variables for estimation of ||r||:
        betadd: float
        thetatilde: float
        rhodold: float
        betad: float
        tautildeold: float
        d: float
        # Variables for estimation of ||A|| and cond(A)
        normA2: float
        maxrbar: float
        minrbar: float
        normA: float
        condA: float
        normx: float

        # Variables for use in stopping rules
        normar: float
        normr: float
        # Reason for stopping
        istop: int

    # more often than not, the matvec is defined after the LSMR
    # solver has been constructed. So it's part of the run()
    # function, not the LSMR constructor.
    def run(vecmat, b, *vecmat_args, damp=0.0):
        def vecmat_noargs(v):
            return vecmat(v, *vecmat_args)

        (ncols,) = func.eval_shape(vecmat, b, *vecmat_args).shape

        state, normb, matvec_noargs = init(vecmat_noargs, b, ncols=ncols)
        step_fun = make_step(matvec_noargs, normb=normb, damp=damp)
        cond_fun = make_cond_fun()
        state = while_loop(cond_fun, step_fun, state)
        stats_ = stats(state)
        return state.x, stats_

    def init(vecmat, b, ncols: int):
        normb = linalg.vector_norm(b)
        x = np.zeros(ncols, dtype=b.dtype)
        beta = normb

        u = b
        u = u / np.where(beta > 0, beta, 1.0)

        v, matvec = func.vjp(vecmat, u)
        alpha = linalg.vector_norm(v)
        v = v / np.where(alpha > 0, alpha, 1)
        v = np.where(beta == 0, np.zeros_like(v), v)
        alpha = np.where(beta == 0, np.zeros_like(alpha), alpha)

        # Initialize variables for 1st iteration.

        zetabar = alpha * beta
        alphabar = alpha
        rho = 1.0
        rhobar = 1.0
        cbar = 1.0
        sbar = 0.0

        h = v
        hbar = np.zeros(ncols, dtype=b.dtype)

        # Initialize variables for estimation of ||r||.

        betadd = beta
        betad = 0.0
        rhodold = 1.0
        tautildeold = 0.0
        thetatilde = 0.0
        zeta = 0.0
        d = 0.0

        # Initialize variables for estimation of ||A|| and cond(A)

        normA2 = alpha * alpha
        maxrbar = 0.0
        minrbar = 1e10
        normA = np.sqrt(normA2)
        condA = 1.0
        normx = 0.0

        # Items for use in stopping rules, normb set earlier
        normr = beta

        # Reverse the order here from the original matlab code because
        # there was an error on return when arnorm==0
        normar = alpha * beta

        # Main iteration loop.
        state = State(  # type: ignore
            itn=0,
            alpha=alpha,
            u=u,
            v=v,
            alphabar=alphabar,
            rho=rho,
            rhobar=rhobar,
            zeta=zeta,
            sbar=sbar,
            cbar=cbar,
            zetabar=zetabar,
            hbar=hbar,
            h=h,
            x=x,
            betadd=betadd,
            thetatilde=thetatilde,
            rhodold=rhodold,
            betad=betad,
            tautildeold=tautildeold,
            d=d,
            normA2=normA2,
            maxrbar=maxrbar,
            minrbar=minrbar,
            normar=normar,
            normr=normr,
            normA=normA,
            condA=condA,
            normx=normx,
            istop=0,
        )
        state = tree.tree_map(np.asarray, state)
        return state, normb, lambda *a: matvec(*a)[0]

    def make_step(matvec, normb: float, damp) -> Callable:
        def step(state: State) -> State:
            # Perform the next step of the bidiagonalization

            Av, A_t = func.vjp(matvec, state.v)
            u = Av - state.alpha * state.u
            beta = linalg.vector_norm(u)

            u = u / np.where(beta > 0, beta, 1.0)
            v = A_t(u)[0] - beta * state.v
            alpha = linalg.vector_norm(v)
            v = v / np.where(alpha > 0, alpha, 1)

            # Construct rotation Qhat_{k,2k+1}.

            chat, shat, alphahat = _sym_ortho(state.alphabar, damp)

            # Use a plane rotation (Q_i) to turn B_i to R_i

            rhoold = state.rho
            c, s, rho = _sym_ortho(alphahat, beta)
            thetanew = s * alpha
            alphabar = c * alpha

            # Use a plane rotation (Qbar_i) to turn R_i^T to R_i^bar

            rhobarold = state.rhobar
            zetaold = state.zeta
            thetabar = state.sbar * rho
            rhotemp = state.cbar * rho
            cbar, sbar, rhobar = _sym_ortho(rhotemp, thetanew)
            zeta = cbar * state.zetabar
            zetabar = -sbar * state.zetabar

            # Update h, h_hat, x.

            hbar = state.h - state.hbar * (thetabar * rho / (rhoold * rhobarold))
            x = state.x + (zeta / (rho * rhobar)) * hbar
            h = v - state.h * (thetanew / rho)

            # Estimate of ||r||.

            # Apply rotation Qhat_{k,2k+1}.
            betaacute = chat * state.betadd
            betacheck = -shat * state.betadd

            # Apply rotation Q_{k,k+1}.
            betahat = c * betaacute
            betadd = -s * betaacute

            # Apply rotation Qtilde_{k-1}.

            thetatildeold = state.thetatilde
            ctildeold, stildeold, rhotildeold = _sym_ortho(state.rhodold, thetabar)
            thetatilde = stildeold * rhobar
            rhodold = ctildeold * rhobar
            betad = -stildeold * state.betad + ctildeold * betahat

            tautildeold = (zetaold - thetatildeold * state.tautildeold) / rhotildeold
            taud = (zeta - thetatilde * tautildeold) / rhodold
            d = state.d + betacheck * betacheck
            normr = np.sqrt(d + (betad - taud) ** 2 + betadd * betadd)

            # Estimate ||A||.
            normA2 = state.normA2 + beta * beta
            normA = np.sqrt(normA2)
            normA2 = normA2 + alpha * alpha

            # Estimate cond(A).
            maxrbar = np.elementwise_max(state.maxrbar, rhobarold)
            minrbar = np.where(
                state.itn > 1,
                np.elementwise_min(state.minrbar, rhobarold),
                state.minrbar,
            )
            condA = np.elementwise_max(maxrbar, rhotemp) / np.elementwise_min(
                minrbar, rhotemp
            )

            # Compute norms for convergence testing.
            normar = np.abs(zetabar)
            normx = linalg.vector_norm(x)

            # Check whether we should stop
            itn = state.itn + 1
            test1 = normr / normb
            z = normA * normr
            z_safe = np.where(z != 0, z, 1.0)
            test2 = np.where(z != 0, normar / z_safe, _LARGE_VALUE)
            test3 = 1 / condA
            t1 = test1 / (1 + normA * normx / normb)
            rtol = btol + atol * normA * normx / normb

            # Early exits
            istop = 0
            istop = np.where(normar == 0, 9, istop)
            istop = np.where(normb == 0, 8, istop)
            istop = np.where(itn >= maxiter, 7, istop)
            istop = np.where(1 + test3 <= 1, 6, istop)
            istop = np.where(1 + test2 <= 1, 5, istop)
            istop = np.where(1 + t1 <= 1, 4, istop)
            istop = np.where(test3 <= ctol, 3, istop)
            istop = np.where(test2 <= atol, 2, istop)
            istop = np.where(test1 <= rtol, 1, istop)

            return State(  # type: ignore
                itn=itn,
                alpha=alpha,
                u=u,
                v=v,
                alphabar=alphabar,
                rho=rho,
                rhobar=rhobar,
                zeta=zeta,
                sbar=sbar,
                cbar=cbar,
                zetabar=zetabar,
                hbar=hbar,
                h=h,
                x=x,
                betadd=betadd,
                thetatilde=thetatilde,
                rhodold=rhodold,
                betad=betad,
                tautildeold=tautildeold,
                d=d,
                normA2=normA2,
                maxrbar=maxrbar,
                minrbar=minrbar,
                normar=normar,
                normr=normr,
                normA=normA,
                condA=condA,
                normx=normx,
                istop=istop,
            )

        return step

    def make_cond_fun() -> Callable:
        def cond(state):
            state_flat, _ = tree.ravel_pytree(state)
            no_nans = np.logical_not(np.any(np.isnan(state_flat)))
            proceed = np.where(state.istop == 0, True, False)
            return np.logical_and(proceed, no_nans)

        return cond

    def stats(state: State) -> dict:
        return {
            "iteration_count": state.itn,
            "norm_residual": state.normr,
            "norm_At_residual": state.normar,
            "norm_A": state.normA,
            "cond_A": state.condA,
            "norm_x": state.normx,
            "istop": state.istop,
        }

    return run


def _sym_ortho(a, b):
    """Stable implementation of Givens rotation. Like in Scipy."""
    idx = 3  # The "else" branch.
    idx = np.where(np.abs(b) > np.abs(a), 2, idx)
    idx = np.where(a == 0, 1, idx)
    idx = np.where(b == 0, 0, idx)

    branches = [_sym_ortho_0, _sym_ortho_1, _sym_ortho_2, _sym_ortho_3]
    return control_flow.switch(idx, branches, a, b)


def _sym_ortho_0(a, _b):
    zero = np.zeros((), dtype=a.dtype)
    return np.sign(a), zero, np.abs(a)


def _sym_ortho_1(_a, b):
    zero = np.zeros((), dtype=b.dtype)
    return zero, np.sign(b), np.abs(b)


def _sym_ortho_2(a, b):
    tau = a / b
    s = np.sign(b) / np.sqrt(1 + tau * tau)
    c = s * tau
    r = b / s
    return c, s, r


def _sym_ortho_3(a, b):
    tau = b / a
    c = np.sign(a) / np.sqrt(1 + tau * tau)
    s = c * tau
    r = a / c
    return c, s, r



"""Test the tri-diagonalisation."""

from matfree import decomp, test_util
from matfree.backend import linalg, np, testing


@testing.parametrize("reortho", ["full", "none"])
@testing.parametrize("ndim", [12])
def test_full_rank_reconstruction_is_exact(reortho, ndim):
    # Set up a test-matrix and an initial vector
    eigvals = np.arange(1.0, 2.0, step=1 / ndim)
    matrix = test_util.symmetric_matrix_from_eigenvalues(eigvals)
    vector = np.flip(np.arange(1.0, 1.0 + len(eigvals)))

    # Run Lanczos approximation
    algorithm = decomp.tridiag_sym(ndim, reortho=reortho, materialize=True)
    Q, T, *_ = algorithm(lambda s, p: p @ s, vector, matrix)

    # Reconstruct the original matrix from the full-num_matvecs approximation
    matrix_reconstructed = Q @ T @ Q.T

    if reortho == "full":
        tols = {"atol": 1e-5, "rtol": 1e-5}
    else:
        tols = {"atol": 1e-1, "rtol": 1e-1}

    # Assert the reconstruction was "exact"
    assert np.allclose(matrix_reconstructed, matrix, **tols)

    # Assert all vectors are orthogonal
    test_util.assert_columns_orthonormal(Q)
    test_util.assert_columns_orthonormal(Q.T)


# anything 0 <= k < n works; k=n is full reconstruction
# and the (q, b) values become meaningless
@testing.parametrize("num_matvecs", [1, 5, 11])
@testing.parametrize("ndim", [12])
@testing.parametrize("reortho", ["full", "none"])
def test_mid_rank_reconstruction_satisfies_decomposition(ndim, num_matvecs, reortho):
    # Set up a test-matrix and an initial vector
    eigvals = np.arange(1.0, 2.0, step=1 / ndim)
    matrix = test_util.symmetric_matrix_from_eigenvalues(eigvals)
    vector = np.flip(np.arange(1.0, 1.0 + len(eigvals)))

    # Run Lanczos approximation
    algorithm = decomp.tridiag_sym(num_matvecs, reortho=reortho, materialize=True)
    Q, T, q, _n = algorithm(lambda s, p: p @ s, vector, matrix)

    # Verify the decomposition
    e_K = np.eye(num_matvecs)[-1]
    test_util.assert_allclose(matrix @ Q, Q @ T + linalg.outer(q, e_K))



"""Stochastic estimation of traces, diagonals, and more."""

from matfree.backend import func, linalg, np, prng, tree
from matfree.backend.typing import Callable


def estimator(integrand: Callable, /, sampler: Callable) -> Callable:
    """Construct a stochastic trace-/diagonal-estimator.

    Parameters
    ----------
    integrand
        The integrand function. For example, the return-value of
        [integrand_trace][matfree.stochtrace.integrand_trace].
        But any other integrand works, too.
    sampler
        The sample function. Usually, either
        [sampler_normal][matfree.stochtrace.sampler_normal] or
        [sampler_rademacher][matfree.stochtrace.sampler_rademacher].

    Returns
    -------
    estimate
        A function that maps a random key to an estimate.
        This function can be compiled, vectorised, differentiated,
        or looped over as the user desires.

    """

    def estimate(matvecs, key, *parameters):
        samples = sampler(key)
        Qs = func.vmap(lambda vec: integrand(matvecs, vec, *parameters))(samples)
        return tree.tree_map(lambda s: np.mean(s, axis=0), Qs)

    return estimate


def integrand_diagonal():
    """Construct the integrand for estimating the diagonal.

    When plugged into the Monte-Carlo estimator,
    the result will be an Array or PyTree of Arrays with the
    same tree-structure as
    ``
    matvec(*args_like)
    ``
    where ``*args_like`` is an argument of the sampler.
    """

    def integrand(matvec, v, *parameters):
        Qv = matvec(v, *parameters)
        v_flat, unflatten = tree.ravel_pytree(v)
        Qv_flat, _unflatten = tree.ravel_pytree(Qv)
        return unflatten(v_flat * Qv_flat)

    return integrand


def integrand_trace():
    """Construct the integrand for estimating the trace."""

    def integrand(matvec, v, *parameters):
        Qv = matvec(v, *parameters)
        v_flat, unflatten = tree.ravel_pytree(v)
        Qv_flat, _unflatten = tree.ravel_pytree(Qv)
        return linalg.inner(v_flat, Qv_flat)

    return integrand


def integrand_trace_and_diagonal():
    """Construct the integrand for estimating the trace and diagonal jointly."""

    def integrand(matvec, v, *parameters):
        Qv = matvec(v, *parameters)
        v_flat, unflatten = tree.ravel_pytree(v)
        Qv_flat, _unflatten = tree.ravel_pytree(Qv)
        trace_form = linalg.inner(v_flat, Qv_flat)
        diagonal_form = unflatten(v_flat * Qv_flat)
        return {"trace": trace_form, "diagonal": diagonal_form}

    return integrand


def integrand_frobeniusnorm_squared():
    """Construct the integrand for estimating the squared Frobenius norm."""

    def integrand(matvec, vec, *parameters):
        x = matvec(vec, *parameters)
        v_flat, unflatten = tree.ravel_pytree(x)
        return linalg.inner(v_flat, v_flat)

    return integrand


def integrand_wrap_moments(integrand, /, moments):
    """Wrap an integrand into another integrand that computes moments.

    Parameters
    ----------
    integrand
        Any integrand function compatible with Hutchinson-style estimation.
    moments
        Any Pytree (tuples, lists, dictionaries) whose leafs that are
        valid inputs to ``lambda m: x**m`` for an array ``x``,
        usually, with data-type ``float``
        (but that depends on the wrapped integrand).
        For example, ``moments=4``, ``moments=(1,2)``,
        or ``moments={"a": 1, "b": 2}``.

    Returns
    -------
    integrand
        An integrand function compatible with Hutchinson-style estimation whose
        output has a PyTree-structure that mirrors the structure of the ``moments``
        argument.

    """

    def integrand_wrapped(vec, *parameters):
        Qs = integrand(vec, *parameters)
        return tree.tree_map(moment_fun, Qs)

    def moment_fun(x, /):
        return tree.tree_map(lambda m: x**m, moments)

    return integrand_wrapped


def sampler_normal(*args_like, num):
    """Construct a function that samples from a standard-normal distribution."""
    return _sampler_from_jax_random(prng.normal, *args_like, num=num)


def sampler_rademacher(*args_like, num):
    """Construct a function that samples from a Rademacher distribution."""
    return _sampler_from_jax_random(prng.rademacher, *args_like, num=num)


def _sampler_from_jax_random(sampler, *args_like, num):
    x_flat, unflatten = tree.ravel_pytree(*args_like)

    def sample(key):
        samples = sampler(key, shape=(num, *x_flat.shape), dtype=x_flat.dtype)
        return func.vmap(unflatten)(samples)

    return sample



"""Tests for Lanczos functionality."""

from matfree import decomp, funm, stochtrace, test_util
from matfree.backend import linalg, np, prng, testing


def make_A(n, num_significant_eigvals):
    """Make a positive definite matrix with certain spectrum."""
    # 'Invent' a spectrum. Use the number of pre-defined eigenvalues.
    d = np.arange(n) / n + 1.0
    d = d.at[num_significant_eigvals:].set(0.001)

    return test_util.symmetric_matrix_from_eigenvalues(d)


@testing.parametrize("n", [200])
@testing.parametrize("num_significant_eigvals", [30])
@testing.parametrize("num_matvecs", [10])
# usually: ~1.5 * num_significant_eigvals.
# But logdet seems to converge sooo much faster.
def test_logdet_spd(n, num_significant_eigvals, num_matvecs):
    """Assert that the log-determinant estimation matches the true log-determinant."""
    A = make_A(n, num_significant_eigvals)

    def matvec(x):
        return {"fx": A @ x["fx"]}

    key = prng.prng_key(1)
    args_like = {"fx": np.ones((n,), dtype=float)}
    sampler = stochtrace.sampler_normal(args_like, num=10)
    tridiag_sym = decomp.tridiag_sym(num_matvecs, materialize=True)
    integrand = funm.integrand_funm_sym_logdet(tridiag_sym)
    estimate = stochtrace.estimator(integrand, sampler)
    received = estimate(matvec, key)

    expected = linalg.slogdet(A)[1]
    print_if_assert_fails = ("error", np.abs(received - expected), "target:", expected)
    assert np.allclose(received, expected, atol=1e-2, rtol=1e-2), print_if_assert_fails


@testing.parametrize("n", [50])
# usually: ~1.5 * num_significant_eigvals.
# But logdet seems to converge sooo much faster.
def test_logdet_spd_exact_for_full_num_matvecs_lanczos(n):
    r"""Computing v^\top f(A) v with max-order Lanczos should be exact for _any_ v."""
    # Construct a (numerically nice) matrix
    eigvals = np.arange(1.0, 1.0 + n, step=1.0)
    A = test_util.symmetric_matrix_from_eigenvalues(eigvals)

    # Set up max-num_matvecs Lanczos approximation inside SLQ for the matrix-logarithm
    num_matvecs = n - 1
    tridiag_sym = decomp.tridiag_sym(num_matvecs, materialize=True)
    integrand = funm.integrand_funm_sym_logdet(tridiag_sym)

    # Construct a vector without that does not have expected 2-norm equal to "dim"
    x = prng.normal(prng.prng_key(seed=1), shape=(n,)) + 10

    # Compute v^\top @ log(A) @ v via Lanczos
    received = integrand(lambda v: A @ v, x)

    # Compute the "true" value of v^\top @ log(A) @ v via eigenvalues
    eigvals, eigvecs = linalg.eigh(A)
    logA = eigvecs @ linalg.diagonal_matrix(np.log(eigvals)) @ eigvecs.T
    expected = x.T @ logA @ x

    # They should be identical
    assert np.allclose(received, expected)



from matfree import decomp, test_util
from matfree.backend import config, func, linalg, np, prng, testing


@testing.parametrize("nrows", [3])
@testing.parametrize("num_matvecs", [2])
@testing.parametrize("reortho", ["none", "full"])
@testing.parametrize("dtype", [float])
def test_adjoint_matches_jax_dot_vjp(nrows, num_matvecs, reortho, dtype):
    # Create a matrix and a direction as a test-case
    A = prng.normal(prng.prng_key(1), shape=(nrows, nrows), dtype=dtype)
    v = prng.normal(prng.prng_key(2), shape=(nrows,), dtype=dtype)

    # Set up the algorithms
    algorithm_autodiff = decomp.hessenberg(
        num_matvecs, reortho=reortho, custom_vjp=False
    )
    algorithm_adjoint = decomp.hessenberg(num_matvecs, reortho=reortho, custom_vjp=True)

    # Forward pass
    algorithm_autodiff = func.partial(algorithm_autodiff, lambda s, p: p @ s)
    algorithm_adjoint = func.partial(algorithm_adjoint, lambda s, p: p @ s)
    fwd_output, vjp_autodiff = func.vjp(algorithm_autodiff, v, A)
    _fwd_output, vjp_adjoint = func.vjp(algorithm_adjoint, v, A)

    # Random input gradients (no sparsity at all)
    vjp_input = test_util.tree_random_like(prng.prng_key(3), fwd_output)

    # Call the auto-diff VJP
    dv_autodiff, dp_autodiff = vjp_autodiff(vjp_input)
    dv_adjoint, dp_adjoint = vjp_adjoint(vjp_input)

    # Assert gradients match
    test_util.assert_allclose(dv_adjoint, dv_autodiff)
    test_util.assert_allclose(dp_adjoint, dp_autodiff)

    # Assert that the values are only similar, not identical
    assert not np.all(dv_adjoint == dv_autodiff)
    assert not np.all(dp_adjoint == dp_autodiff)


@testing.parametrize("nrows", [15])
@testing.parametrize("num_matvecs", [10])
@testing.parametrize("reortho", ["full"])
def test_adjoint_matches_jax_dot_vjp_hilbert_matrix_and_full_reortho(
    nrows, num_matvecs, reortho
):
    config.update("jax_enable_x64", True)
    # Create a matrix and a direction as a test-case
    A = _lower(linalg.hilbert(nrows))
    v = prng.normal(prng.prng_key(2), shape=(nrows,), dtype=A.dtype)

    def matvec(s, p):
        return (p + p.T) @ s

    # Set up the algorithms
    algorithm_autodiff = decomp.hessenberg(
        num_matvecs, reortho=reortho, custom_vjp=False
    )
    algorithm_adjoint = decomp.hessenberg(num_matvecs, reortho=reortho, custom_vjp=True)

    # Forward pass
    algorithm_autodiff = func.partial(algorithm_autodiff, matvec)
    algorithm_adjoint = func.partial(algorithm_adjoint, matvec)
    fwd_output, vjp_autodiff = func.vjp(algorithm_autodiff, v, A)
    _fwd_output, vjp_adjoint = func.vjp(algorithm_adjoint, v, A)

    # Random input gradients (no sparsity at all)
    vjp_input = test_util.tree_random_like(prng.prng_key(3), fwd_output)

    # Call the auto-diff VJP
    dv_autodiff, dp_autodiff = vjp_autodiff(vjp_input)
    dv_adjoint, dp_adjoint = vjp_adjoint(vjp_input)

    # Assert gradients match
    test_util.assert_allclose(dv_adjoint, dv_autodiff)
    test_util.assert_allclose(dp_adjoint, dp_autodiff)

    # Assert that the values are only similar, not identical
    assert not np.all(dv_adjoint == dv_autodiff)
    assert not np.all(dp_adjoint == dp_autodiff)

    config.update("jax_enable_x64", False)


def _lower(m):
    m_tril = np.tril(m)
    return m_tril - 0.5 * linalg.diagonal_matrix(linalg.diagonal(m_tril))


@testing.parametrize("reortho_wrong", [True, "full_with_sparsity", "None"])
def test_raises_type_error_for_wrong_reorthogonalisation_flag(reortho_wrong):
    # Create a matrix and a direction as a test-case

    # Set up the algorithms
    with testing.raises(TypeError, match="Unexpected input"):
        _ = decomp.hessenberg(1, reortho=reortho_wrong)



"""Test utilities."""

from matfree.backend import linalg, np, prng, tree


def symmetric_matrix_from_eigenvalues(eigvals, /):
    """Generate a symmetric matrix with prescribed eigenvalues."""
    (n,) = eigvals.shape

    # Need _some_ matrix to start with
    A = np.reshape(np.arange(1.0, n**2 + 1.0), (n, n))
    A = A / linalg.matrix_norm(A, which="fro")
    X = A.T @ A + np.eye(n)

    # QR decompose. We need the orthogonal matrix.
    # Treat Q as a stack of eigenvectors.
    Q, _R = linalg.qr_reduced(X)

    # Treat Q as eigenvectors, and 'D' as eigenvalues.
    # return Q D Q.T.
    # This matrix will be dense, symmetric, and have a given spectrum.
    return Q @ (eigvals[:, None] * Q.T)


def asymmetric_matrix_from_singular_values(vals, /, nrows, ncols):
    """Generate an asymmetric matrix with specific singular values."""
    A = np.reshape(np.arange(1.0, nrows * ncols + 1.0), (nrows, ncols))
    A /= nrows * ncols
    U, S, Vt = linalg.svd(A, full_matrices=False)
    return U @ linalg.diagonal(vals) @ Vt


def to_dense_bidiag(d, e, /, offset=1):
    """Materialize a bidiagonal matrix."""
    diag = linalg.diagonal_matrix(d)
    offdiag = linalg.diagonal_matrix(e, offset=offset)
    return diag + offdiag


def to_dense_tridiag_sym(d, e, /):
    """Materialize a symmetric tridiagonal matrix."""
    diag = linalg.diagonal_matrix(d)
    offdiag1 = linalg.diagonal_matrix(e, offset=1)
    offdiag2 = linalg.diagonal_matrix(e, offset=-1)
    return diag + offdiag1 + offdiag2


def tree_random_like(key, pytree, *, generate_func=prng.normal):
    """Fill a tree with random values."""
    flat, unflatten = tree.ravel_pytree(pytree)
    flat_like = generate_func(key, shape=flat.shape, dtype=flat.dtype)
    return unflatten(flat_like)


def assert_columns_orthonormal(Q, /):
    """Assert that the columns in a matrix are orthonormal."""
    eye_like = Q.T @ Q
    ref = np.eye(len(eye_like))
    assert_allclose(eye_like, ref)


def assert_allclose(a, b, /):
    """Assert that two arrays are close.

    This function uses a different default tolerance to
    jax.numpy.allclose. Instead of fixing values, the tolerance
    depends on the floating-point precision of the input variables.
    """
    a = np.asarray(a)
    b = np.asarray(b)
    tol = np.sqrt(np.finfo_eps(np.dtype(b)))

    # For double precision sqrt(eps) is very tight...
    if tol < 1e-7:
        tol *= 100
    assert np.allclose(a, b, atol=tol, rtol=tol)



"""Control flow."""

import jax

# API follows JAX, but we liberally work with positional- and keyword-only arguments
# We also rename some arguments for improved consistency:
# For example, we always use 'body_fun' and 'init_val',
#  even though jax.lax.scan uses 'f' and 'init'.


def scan(body_fun, init_val, /, xs, *, reverse=False):
    return jax.lax.scan(body_fun, init=init_val, xs=xs, reverse=reverse)


def cond(pred, /, true_fun, false_fun, *operands):
    return jax.lax.cond(pred, true_fun, false_fun, *operands)


def fori_loop(lower, upper, body_fun, init_val):
    return jax.lax.fori_loop(lower, upper, body_fun, init_val)


def while_loop(cond_fun, body_fun, init_val):
    return jax.lax.while_loop(cond_fun, body_fun, init_val)


def switch(index, branches, *operands):
    return jax.lax.switch(index, branches, *operands)



"""Test the Golub-Kahan-Lanczos bi-diagonalisation with full re-orthogonalisation."""

from matfree import decomp, test_util
from matfree.backend import linalg, np, prng, testing


def make_A(nrows, ncols, num_significant_singular_vals):
    """Make a positive definite matrix with certain spectrum."""
    # 'Invent' a spectrum. Use the number of pre-defined eigenvalues.
    n = min(nrows, ncols)
    d = np.arange(n) + 10.0
    d = d.at[num_significant_singular_vals:].set(0.001)
    return test_util.asymmetric_matrix_from_singular_values(d, nrows=nrows, ncols=ncols)


@testing.parametrize("nrows", [50])
@testing.parametrize("ncols", [49])
@testing.parametrize("num_significant_singular_vals", [4])
@testing.parametrize("num_matvecs", [6])  # ~1.5 * num_significant_eigvals
def test_bidiag_decomposition_is_satisfied(
    nrows, ncols, num_significant_singular_vals, num_matvecs
):
    """Test that Lanczos tridiagonalisation yields an orthogonal-tridiagonal decomp."""
    A = make_A(nrows, ncols, num_significant_singular_vals)
    key = prng.prng_key(1)
    v0 = prng.normal(key, shape=(ncols,))

    algorithm = decomp.bidiag(num_matvecs, materialize=True)
    (U, V), B, res, ln = algorithm(lambda v: A @ v, v0)

    test_util.assert_columns_orthonormal(U)
    test_util.assert_columns_orthonormal(V)

    em = np.eye(num_matvecs)[:, -1]
    test_util.assert_allclose(A @ V, U @ B)
    test_util.assert_allclose(A.T @ U, V @ B.T + linalg.outer(res, em))
    test_util.assert_allclose(1.0 / linalg.vector_norm(v0), ln)



"""Test stochastic Lanczos quadrature for log-determinants of matrix-products."""

from matfree import decomp, funm, stochtrace, test_util
from matfree.backend import linalg, np, prng, testing


def make_A(nrows, ncols, num_significant_singular_vals):
    """Make a positive definite matrix with certain spectrum."""
    # 'Invent' a spectrum. Use the number of pre-defined eigenvalues.
    n = min(nrows, ncols)
    d = np.arange(n) + 1.0
    d = d.at[num_significant_singular_vals:].set(0.001)
    return test_util.asymmetric_matrix_from_singular_values(d, nrows=nrows, ncols=ncols)


@testing.parametrize("nrows", [50])
@testing.parametrize("ncols", [30])
@testing.parametrize("num_significant_singular_vals", [30])
@testing.parametrize("num_matvecs", [20])
def test_logdet_product(nrows, ncols, num_significant_singular_vals, num_matvecs):
    """Assert that logdet_product yields an accurate estimate."""
    A = make_A(nrows, ncols, num_significant_singular_vals)
    key = prng.prng_key(3)

    def matvec(x):
        return {"fx": A @ x["fx"]}

    x_like = {"fx": np.ones((ncols,), dtype=float)}
    fun = stochtrace.sampler_rademacher(x_like, num=400)

    bidiag = decomp.bidiag(num_matvecs)
    problem = funm.integrand_funm_product_logdet(bidiag)
    estimate = stochtrace.estimator(problem, fun)
    received = estimate(matvec, key)

    expected = linalg.slogdet(A.T @ A)[1]
    print_if_assert_fails = ("error", np.abs(received - expected), "target:", expected)
    assert np.allclose(received, expected, atol=1e-2, rtol=1e-2), print_if_assert_fails


@testing.parametrize("n", [50])
# usually: ~1.5 * num_significant_eigvals.
# But logdet seems to converge sooo much faster.
def test_logdet_product_exact_for_full_num_matvecs_lanczos(n):
    r"""Computing v^\top f(A^\top @ A) v with max-order Lanczos is exact for _any_ v."""
    # Construct a (numerically nice) matrix
    singular_values = np.sqrt(np.arange(1.0, 1.0 + n, step=1.0))
    A = test_util.asymmetric_matrix_from_singular_values(
        singular_values, nrows=n, ncols=n
    )

    # Set up max-num_matvecs Lanczos approximation inside SLQ for the matrix-logarithm
    num_matvecs = n - 1
    bidiag = decomp.bidiag(num_matvecs)
    integrand = funm.integrand_funm_product_logdet(bidiag)

    # Construct a vector without that does not have expected 2-norm equal to "dim"
    x = prng.normal(prng.prng_key(seed=1), shape=(n,)) + 1

    # Compute v^\top @ log(A) @ v via Lanczos
    received = integrand((lambda v: A @ v), x)

    # Compute the "true" value of v^\top @ log(A) @ v via eigenvalues
    eigvals, eigvecs = linalg.eigh(A.T @ A)
    logA = eigvecs @ linalg.diagonal_matrix(np.log(eigvals)) @ eigvecs.T
    expected = x.T @ logA @ x

    # They should be identical
    assert np.allclose(received, expected)



r"""Matrix-free implementations of functions of matrices.

This includes matrix-function-vector products

$$
(f, A, v, p) \mapsto f(A(p))v
$$

as well as matrix-function extensions for stochastic trace estimation,
which provide

$$
(f, A, v, p) \mapsto v^\top f(A(p))v.
$$

Plug these integrands into
[matfree.stochtrace.estimator][matfree.stochtrace.estimator].


Examples
--------
>>> import jax.random
>>> import jax.numpy as jnp
>>> from matfree import decomp
>>>
>>> M = jax.random.normal(jax.random.PRNGKey(1), shape=(10, 10))
>>> A = M.T @ M
>>> v = jax.random.normal(jax.random.PRNGKey(2), shape=(10,))
>>>
>>> # Compute a matrix-logarithm with Lanczos' algorithm
>>> matfun = dense_funm_sym_eigh(jnp.log)
>>> tridiag = decomp.tridiag_sym(4)
>>> matfun_vec = funm_lanczos_sym(matfun, tridiag)
>>> fAx = matfun_vec(lambda s: A @ s, v)
>>> print(fAx.shape)
(10,)

"""

from matfree.backend import containers, control_flow, func, linalg, np, tree
from matfree.backend.typing import Array, Callable


def funm_chebyshev(matfun: Callable, num_matvecs: int, matvec: Callable, /) -> Callable:
    """Compute a matrix-function-vector product via Chebyshev's algorithm.

    This function assumes that the **spectrum of the matrix-vector product
    is contained in the interval (-1, 1)**, and that the **matrix-function
    is analytic on this interval**. If this is not the case,
    transform the matrix-vector product and the matrix-function accordingly.
    """
    # Construct nodes
    nodes = _chebyshev_nodes(num_matvecs)
    fx_nodes = matfun(nodes)

    class _ChebyshevState(containers.NamedTuple):
        interpolation: Array
        poly_coefficients: tuple[Array, Array]
        poly_values: tuple[Array, Array]

    def init_func(vec, *parameters):
        # Initialize the scalar recursion
        # (needed to compute the interpolation weights)
        t2_n, t1_n = nodes, np.ones_like(nodes)
        c1 = np.mean(fx_nodes * t1_n)
        c2 = 2 * np.mean(fx_nodes * t2_n)

        # Initialize the vector-valued recursion
        # (this is where the matvec happens)
        t2_x, t1_x = matvec(vec, *parameters), vec
        value = c1 * t1_x + c2 * t2_x
        return _ChebyshevState(value, (t2_n, t1_n), (t2_x, t1_x))

    def recursion_func(val: _ChebyshevState, *parameters) -> _ChebyshevState:
        value, (t2_n, t1_n), (t2_x, t1_x) = val

        # Apply the next scalar recursion and
        # compute the next coefficient
        t2_n, t1_n = 2 * nodes * t2_n - t1_n, t2_n
        c2 = 2 * np.mean(fx_nodes * t2_n)

        # Apply the next matrix-vector product recursion and
        # compute the next interpolation-value
        t2_x, t1_x = 2 * matvec(t2_x, *parameters) - t1_x, t2_x
        value += c2 * t2_x
        return _ChebyshevState(value, (t2_n, t1_n), (t2_x, t1_x))

    def extract_func(val: _ChebyshevState):
        return val.interpolation

    alg = (0, num_matvecs - 1), init_func, recursion_func, extract_func
    return _funm_polyexpand(alg)


def _chebyshev_nodes(n, /):
    k = np.arange(n, step=1.0) + 1
    return np.cos((2 * k - 1) / (2 * n) * np.pi())


def _funm_polyexpand(matrix_poly_alg, /):
    """Compute a matrix-function-vector product via a polynomial expansion."""
    (lower, upper), init_func, step_func, extract_func = matrix_poly_alg

    def matrix_function_vector_product(vec, *parameters):
        final_state = control_flow.fori_loop(
            lower=lower,
            upper=upper,
            body_fun=lambda _i, v: step_func(v, *parameters),
            init_val=init_func(vec, *parameters),
        )
        return extract_func(final_state)

    return matrix_function_vector_product


def funm_lanczos_sym(dense_funm: Callable, tridiag_sym: Callable, /) -> Callable:
    """Implement a matrix-function-vector product via Lanczos' tridiagonalisation.

    This algorithm uses Lanczos' tridiagonalisation
    and therefore applies only to symmetric matrices.

    Parameters
    ----------
    dense_funm
        An implementation of a function of a dense matrix.
        For example, the output of
        [funm.dense_funm_sym_eigh][matfree.funm.dense_funm_sym_eigh]
        [funm.dense_funm_schur][matfree.funm.dense_funm_schur]
    tridiag_sym
        An implementation of tridiagonalisation.
        E.g., the output of
        [decomp.tridiag_sym][matfree.decomp.tridiag_sym].
    """

    def estimate(matvec: Callable, vec, *parameters):
        length = linalg.vector_norm(vec)
        vec /= length
        Q, matrix, *_ = tridiag_sym(matvec, vec, *parameters)

        funm = dense_funm(matrix)
        e1 = np.eye(len(matrix))[0, :]
        return length * (Q @ funm @ e1)

    return estimate


def funm_arnoldi(dense_funm: Callable, hessenberg: Callable, /) -> Callable:
    """Implement a matrix-function-vector product via the Arnoldi iteration.

    This algorithm uses the Arnoldi iteration
    and therefore applies only to all square matrices.

    Parameters
    ----------
    dense_funm
        An implementation of a function of a dense matrix.
        For example, the output of
        [funm.dense_funm_sym_eigh][matfree.funm.dense_funm_sym_eigh]
        [funm.dense_funm_schur][matfree.funm.dense_funm_schur]
    hessenberg
        An implementation of Hessenberg-factorisation.
        E.g., the output of
        [decomp.hessenberg][matfree.decomp.hessenberg].
    """

    def estimate(matvec: Callable, vec, *parameters):
        length = linalg.vector_norm(vec)
        vec /= length
        basis, matrix, *_ = hessenberg(matvec, vec, *parameters)

        funm = dense_funm(matrix)
        e1 = np.eye(len(matrix))[0, :]
        return length * (basis @ funm @ e1)

    return estimate


def integrand_funm_sym_logdet(tridiag_sym: Callable, /):
    """Construct the integrand for the log-determinant.

    This function assumes a symmetric, positive definite matrix.

    Parameters
    ----------
    tridiag_sym
        An implementation of tridiagonalisation.
        E.g., the output of
        [decomp.tridiag_sym][matfree.decomp.tridiag_sym].

    """
    dense_funm = dense_funm_sym_eigh(np.log)
    return integrand_funm_sym(dense_funm, tridiag_sym)


def integrand_funm_sym(dense_funm, tridiag_sym, /):
    """Construct the integrand for matrix-function-trace estimation.

    This function assumes a symmetric matrix.

    Parameters
    ----------
    dense_funm
        An implementation of a function of a dense matrix.
        For example, the output of
        [funm.dense_funm_sym_eigh][matfree.funm.dense_funm_sym_eigh]
        [funm.dense_funm_schur][matfree.funm.dense_funm_schur]
    tridiag_sym
        An implementation of tridiagonalisation.
        E.g., the output of
        [decomp.tridiag_sym][matfree.decomp.tridiag_sym].

    """

    def quadform(matvec, v0, *parameters):
        v0_flat, v_unflatten = tree.ravel_pytree(v0)
        length = linalg.vector_norm(v0_flat)
        v0_flat /= length

        def matvec_flat(v_flat, *p):
            v = v_unflatten(v_flat)
            Av = matvec(v, *p)
            flat, unflatten = tree.ravel_pytree(Av)
            return flat

        _, dense, *_ = tridiag_sym(matvec_flat, v0_flat, *parameters)

        fA = dense_funm(dense)
        e1 = np.eye(len(fA))[0, :]
        return length**2 * linalg.inner(e1, fA @ e1)

    return quadform


def integrand_funm_product_logdet(bidiag: Callable, /):
    r"""Construct the integrand for the log-determinant of a matrix-product.

    Here, "product" refers to $X = A^\top A$.
    """
    dense_funm = dense_funm_product_svd(np.log)
    return integrand_funm_product(dense_funm, bidiag)


def integrand_funm_product_schatten_norm(power, bidiag: Callable, /):
    r"""Construct the integrand for the $p$-th power of the Schatten-p norm."""

    def matfun(x):
        """Matrix-function for Schatten-p norms."""
        return x ** (power / 2)

    dense_funm = dense_funm_product_svd(matfun)
    return integrand_funm_product(dense_funm, bidiag)


def integrand_funm_product(dense_funm, algorithm, /):
    r"""Construct the integrand for matrix-function-trace estimation.

    Instead of the trace of a function of a matrix,
    compute the trace of a function of the product of matrices.
    Here, "product" refers to $X = A^\top A$.
    """

    def quadform(matvec, v0, *parameters):
        v0_flat, v_unflatten = tree.ravel_pytree(v0)
        length = linalg.vector_norm(v0_flat)
        v0_flat /= length

        def matvec_flat(v_flat, *p):
            v = v_unflatten(v_flat)
            Av = matvec(v, *p)
            flat, _unflatten = tree.ravel_pytree(Av)
            return flat

        # Decompose into orthogonal-bidiag-orthogonal
        _, B, *_ = algorithm(matvec_flat, v0_flat, *parameters)

        # Evaluate matfun
        fA = dense_funm(B)
        e1 = np.eye(len(fA))[0, :]
        return length**2 * linalg.inner(e1, fA @ e1)

    return quadform


def dense_funm_product_svd(matfun):
    """Implement dense matrix-functions of a product of matrices via SVDs."""

    def dense_funm(matrix, /):
        # Compute SVD of factorisation
        _, S, Vt = linalg.svd(matrix, full_matrices=False)

        # Since Q orthogonal (orthonormal) to v0, Q v = Q[0],
        # and therefore (Q v)^T f(D) (Qv) = Q[0] * f(diag) * Q[0]
        eigvals, eigvecs = S**2, Vt.T
        fx_eigvals = func.vmap(matfun)(eigvals)
        return eigvecs @ (fx_eigvals[:, None] * eigvecs.T)

    return dense_funm


def dense_funm_sym_eigh(matfun):
    """Implement dense matrix-functions via symmetric eigendecompositions.

    Use it to construct one of the matrix-free matrix-function implementations,
    e.g. [matfree.funm.funm_lanczos_sym][matfree.funm.funm_lanczos_sym].
    """

    def fun(dense_matrix):
        eigvals, eigvecs = linalg.eigh(dense_matrix)
        fx_eigvals = func.vmap(matfun)(eigvals)
        return eigvecs @ linalg.diagonal(fx_eigvals) @ eigvecs.T

    return fun


def dense_funm_schur(matfun):
    """Implement dense matrix-functions via symmetric Schur decompositions.

    Use it to construct one of the matrix-free matrix-function implementations,
    e.g. [matfree.funm.funm_lanczos_sym][matfree.funm.funm_lanczos_sym].
    """

    def fun(dense_matrix):
        return linalg.funm_schur(dense_matrix, matfun)

    return fun


def dense_funm_pade_exp():
    """Implement dense matrix-exponentials using a Pade approximation.

    Use it to construct one of the matrix-free matrix-function implementations,
    e.g. [matfree.funm.funm_arnoldi][matfree.funm.funm_arnoldi].
    """

    def fun(dense_matrix):
        return linalg.funm_pade_exp(dense_matrix)

    return fun



"""Matrix-free eigenvalue and singular-value analysis."""

from matfree.backend import func, linalg, tree
from matfree.backend.typing import Array, Callable


def svd_partial(bidiag: Callable) -> Callable:
    """Partial singular value decomposition.

    Combines bidiagonalisation with a full SVD of the (small) bidiagonal matrix.

    Parameters
    ----------
    bidiag:
        An implementation of bidiagonalisation.
        For example, the output of
        [decomp.bidiag][matfree.decomp.bidiag].
        Note how this function assumes that the bidiagonalisation
        materialises the bidiagonal matrix.

    """

    def svd(Av: Callable, v0: Array, *parameters):
        # Flatten in- and outputs
        Av_flat, flattened = _partial_and_flatten_matvec(Av, v0, *parameters)
        (u0_flat, u_unravel), (v0_flat, v_unravel) = flattened

        # Call the flattened SVD
        ut, s, vt = svd_flat(Av_flat, v0_flat)

        # Unravel the singular vectors
        ut_tree = func.vmap(u_unravel)(ut)
        vt_tree = func.vmap(v_unravel)(vt)
        return ut_tree, s, vt_tree

    def svd_flat(Av: Callable, v0: Array):
        # Factorise the matrix
        (u, v), B, *_ = bidiag(Av, v0)

        # Compute SVD of factorisation
        U, S, Vt = linalg.svd(B, full_matrices=False)

        # Combine orthogonal transformations
        return (u @ U).T, S, Vt @ v.T

    return svd


def eigh_partial(tridiag_sym: Callable) -> Callable:
    """Partial symmetric/Hermitian eigenvalue decomposition.

    Combines tridiagonalization with a decomposition
    of the (small) tridiagonal matrix.

    Parameters
    ----------
    tridiag_sym:
        An implementation of tridiagonalization.
        For example, the output of
        [decomp.tridiag_sym][matfree.decomp.tridiag_sym].

    """

    def eigh(Av: Callable, v0: Array, *parameters):
        # Flatten in- and outputs
        Av_flat, flattened = _partial_and_flatten_matvec(Av, v0, *parameters)
        _, (v0_flat, v_unravel) = flattened

        # Call the flattened eigh
        vals, vecs = eigh_flat(Av_flat, v0_flat)

        # Unravel the eigenvectors
        vecs = func.vmap(v_unravel)(vecs)
        return vals, vecs

    def eigh_flat(Av: Callable, v0: Array):
        # Factorise the matrix
        Q, H, *_ = tridiag_sym(Av, v0)

        # Compute eigh of factorisation
        vals, vecs = linalg.eigh(H)
        vecs = Q @ vecs
        return vals, vecs.T

    return eigh


def eig_partial(hessenberg: Callable) -> Callable:
    """Partial eigenvalue decomposition.

    Combines Hessenberg factorisation with a decomposition
    of the (small) Hessenberg matrix.

    Parameters
    ----------
    hessenberg:
        An implementation of Hessenberg factorisation.
        For example, the output of
        [decomp.hessenberg][matfree.decomp.hessenberg].

    """

    def eig(Av: Callable, v0: Array, *parameters):
        # Flatten in- and outputs
        Av_flat, flattened = _partial_and_flatten_matvec(Av, v0, *parameters)
        _, (v0_flat, v_unravel) = flattened

        # Call the flattened eig
        vals, vecs = eig_flat(Av_flat, v0_flat)

        # Unravel the eigenvectors
        vecs = func.vmap(v_unravel)(vecs)
        return vals, vecs

    def eig_flat(Av: Callable, v0: Array):
        # Factorise the matrix
        Q, H, *_ = hessenberg(Av, v0)

        # Compute eig of factorisation
        vals, vecs = linalg.eig(H)
        vecs = Q @ vecs
        return vals, vecs.T

    return eig


def _partial_and_flatten_matvec(Av, v0, *parameters):
    # Evaluate the output shape of Av and flatten
    u0 = func.eval_shape(lambda z: Av(z, *parameters), v0)

    # Flatten in- and outputs
    u0_flat, u_unravel = func.eval_shape(tree.ravel_pytree, u0)
    v0_flat, v_unravel = tree.ravel_pytree(v0)

    def Av_flat(v_flat):
        """Evaluate a flattened matvec."""
        result = Av(v_unravel(v_flat), *parameters)
        result_flat, _ = tree.ravel_pytree(result)
        return result_flat

    return Av_flat, ((u0_flat, u_unravel), (v0_flat, v_unravel))



"""Implement vector calculus in linear complexity.

Implementing vector calculus with conventional
algorithmic differentiation can be inefficient.
For example, computing the divergence of a
vector field requires computing the trace of a Jacobian.
The divergence of a vector field is
important when evaluating Laplacians of scalar functions.

Here is how we can implement divergences and
Laplacians without forming full Jacobian matrices:
"""

import jax
import jax.numpy as jnp

from matfree import stochtrace

# ## Divergences and Laplacians
#
# The divergence of a vector field is the trace of its Jacobian.
# The conventional implementation would look like this:


def divergence_dense(vf):
    """Compute the divergence of a vector field."""

    def div_fn(x):
        J = jax.jacfwd(vf)
        return jnp.trace(J(x))

    return div_fn


# This implementation computes the divergence of a vector field:


def fun(x):
    """Evaluate a scalar valued function."""
    return jnp.dot(x, x) ** 2


x0 = jnp.arange(1.0, 4.0)
gradient = jax.grad(fun)
laplacian = divergence_dense(gradient)
print(jax.hessian(fun)(x0))
print(laplacian(x0))


# But the implementation above requires $O(d^2)$ storage
# because it evaluates the dense Jacobian.
# This is problematic for high-dimensional problems.
#
# ## Matrix-free implementation
#
# If we have access to Jacobian-vector products (which we usually do),
# we can use matrix-free trace estimation
# to approximate divergences and Laplacians without forming full Jacobians:
#


def divergence_matfree(vf, /, *, num):
    """Compute the divergence with Hutchinson's estimator."""

    def divergence(k, x):
        _fx, jvp = jax.linearize(vf, x)
        integrand_laplacian = stochtrace.integrand_trace()
        normal = stochtrace.sampler_normal(x, num=num)
        estimator = stochtrace.estimator(integrand_laplacian, sampler=normal)
        return estimator(jvp, k)

    return divergence


# -
# The difference to the "naive" implementation is that the implicit one
# does not form dense Jacobians. It requires $O(d)$ memory and
# $O(d N)$ operations (for $N$ Monte-Carlo samples).
# For large-scale problems, it may be the only way of computing Laplacians reliably.

laplacian_matfree = divergence_matfree(gradient, num=10_000)
print(laplacian(x0))
print(laplacian_matfree(jax.random.PRNGKey(1), x0))


# In summary, compute matrix-free linear algebra
# and algorithmic differentiation to implement vector calculus.
#
# ## Diagonals of Jacobians
#
# If we replace trace estimation with diagonal estimation,
# we can compute the diagonal of Jacobian matrices in
# $O(d)$ memory and $O(dN)$ operations.



"""Tests for eigenvalue functionality."""

from matfree import decomp, eig, test_util
from matfree.backend import linalg, np, testing


def test_equal_to_linalg_eigh(nrows=10):
    eigvals = np.arange(1.0, 1.0 + nrows)
    A = test_util.symmetric_matrix_from_eigenvalues(eigvals)
    v0 = np.ones((nrows,))
    num_matvecs = nrows

    hessenberg = decomp.hessenberg(num_matvecs, reortho="full")
    alg = eig.eigh_partial(hessenberg)
    vals, vecs = alg(lambda v, p: p @ v, v0, A)

    S, U = linalg.eigh(A)
    assert np.allclose(vecs.T @ vecs, U @ U.T, atol=1e-5, rtol=1e-5)
    assert np.allclose(vals, S)


@testing.parametrize("nrows", [8])
@testing.parametrize("num_matvecs", [8, 4, 0])
def test_shapes_as_expected_vector(nrows, num_matvecs):
    eigvals = np.arange(1.0, 1.0 + nrows)
    A = test_util.symmetric_matrix_from_eigenvalues(eigvals)
    v0 = np.ones((nrows,))

    tridiag_sym = decomp.tridiag_sym(num_matvecs, reortho="full")
    alg = eig.eigh_partial(tridiag_sym)
    S, U = alg(lambda v, p: p @ v, v0, A)
    assert S.shape == (num_matvecs,)
    assert U.shape == (num_matvecs, nrows)


@testing.parametrize("nrows", [10])
@testing.parametrize("num_matvecs", [0, 2, 3])
def test_shapes_as_expected_lists_tuples(nrows, num_matvecs):
    eigvals = np.arange(1.0, 1.0 + nrows)
    A = test_util.symmetric_matrix_from_eigenvalues(eigvals)
    v0 = np.ones((nrows,))

    # Map Pytrees to Pytrees
    def Av(v: list, p) -> list:
        [(x,)] = v
        return [(p @ x,)]

    tridiag_sym = decomp.tridiag_sym(num_matvecs, reortho="none")
    eigh_partial = eig.eigh_partial(tridiag_sym)
    vals, [(vecs,)] = eigh_partial(Av, [(v0,)], A)

    # Ut inherits pytree-shape from the outputs (list),
    # and Vt inherits pytree-shape from the inputs (tuple)
    assert vecs.shape == (num_matvecs, *v0.shape)
    assert vals.shape == (num_matvecs,)



"""Tests for least-squares functionality."""

from matfree import lstsq, test_util
from matfree.backend import linalg, prng, testing
from matfree.backend.typing import Callable


@testing.case()
def case_lstsq_lsmr() -> Callable:
    return lstsq.lsmr(atol=1e-5, btol=1e-5, ctol=1e-5)


def case_A_shape_wide() -> tuple:
    return 3, 6


def case_A_shape_tall() -> tuple:
    return 6, 3


def case_A_shape_square() -> tuple:
    return 3, 3


@testing.parametrize_with_cases("lstsq_fun", cases=".", prefix="case_lstsq_")
@testing.parametrize_with_cases("A_shape", cases=".", prefix="case_A_shape_")
def test_fwd_matches_numpy_lstsq(lstsq_fun: Callable, A_shape: tuple):
    key = prng.prng_key(1)

    key, subkey = prng.split(key, 2)
    matrix = prng.normal(subkey, shape=A_shape)
    key, subkey = prng.split(key, 2)
    rhs = prng.normal(subkey, shape=(A_shape[0],))

    def vecmat(vector):
        return matrix.T @ vector

    received, _stats = lstsq_fun(vecmat, rhs)
    expected = linalg.lstsq(matrix, rhs)[0]
    test_util.assert_allclose(received, expected)


@testing.parametrize_with_cases("lstsq_fun", cases=".", prefix="case_lstsq_")
@testing.parametrize_with_cases("A_shape", cases=".", prefix="case_A_shape_")
def test_fwd_matches_numpy_lstsq_parametrized(lstsq_fun: Callable, A_shape: tuple):
    key = prng.prng_key(1)

    key, subkey = prng.split(key, 2)
    matrix = prng.normal(subkey, shape=A_shape)
    key, subkey = prng.split(key, 2)
    rhs = prng.normal(subkey, shape=(A_shape[0],))

    def vecmat(vector, A):
        return A.T @ vector

    received, _stats = lstsq_fun(vecmat, rhs, matrix)
    expected = linalg.lstsq(matrix, rhs)[0]
    test_util.assert_allclose(received, expected)


@testing.parametrize_with_cases("lstsq_fun", cases=".", prefix="case_lstsq_")
@testing.parametrize_with_cases("A_shape", cases=".", prefix="case_A_shape_")
def test_fwd_matches_numpy_lstsq_damped(lstsq_fun: Callable, A_shape: tuple):
    key = prng.prng_key(1)

    key, subkey = prng.split(key, 2)
    matrix = prng.normal(subkey, shape=A_shape)
    key, subkey = prng.split(key, 2)
    rhs = prng.normal(subkey, shape=(A_shape[0],))

    def vecmat(vector):
        return matrix.T @ vector

    received, _stats = lstsq_fun(vecmat, rhs, damp=0.0)
    expected = linalg.lstsq(matrix, rhs)[0]
    test_util.assert_allclose(received, expected)



"""Test matrix-function-vector products via the Arnoldi iteration."""

from matfree import decomp, funm
from matfree.backend import np, prng, testing


def case_expm():
    return funm.dense_funm_pade_exp()


@testing.parametrize("reortho", ["full", "none"])
@testing.parametrize_with_cases("dense_funm", cases=".", prefix="case_")
def test_funm_arnoldi_matches_schur_implementation(dense_funm, reortho, n=11):
    """Test matrix-function-vector products via the Arnoldi iteration."""
    # Create a test-problem: matvec, matrix function,
    # vector, and parameters (a matrix).

    matrix = prng.normal(prng.prng_key(1), shape=(n, n))
    v = prng.normal(prng.prng_key(2), shape=(n,))

    # Compute the solution
    expected = dense_funm(matrix) @ v

    # Compute the matrix-function vector product
    arnoldi = decomp.hessenberg((n * 3) // 4, reortho=reortho)
    matfun_vec = funm.funm_arnoldi(dense_funm, arnoldi)
    received = matfun_vec(lambda s, p: p @ s, v, matrix)

    assert np.allclose(expected, received, rtol=1e-1, atol=1e-1)



"""Test the partial Cholesky decompositions."""

from matfree import low_rank, test_util
from matfree.backend import linalg, np, prng, testing
from matfree.backend.typing import Callable


def case_cholesky_partial():
    return low_rank.cholesky_partial


def case_cholesky_partial_pivot():
    return low_rank.cholesky_partial_pivot


@testing.parametrize_with_cases("cholesky", cases=".", prefix="case_cholesky")
def test_full_rank_cholesky_reconstructs_matrix(cholesky, n=5):
    key = prng.prng_key(2)

    cov_eig = 1.0 + prng.uniform(key, shape=(n,), dtype=float)
    cov = test_util.symmetric_matrix_from_eigenvalues(cov_eig)

    approximation, _info = cholesky(lambda i, j: cov[i, j], nrows=n, rank=n)()

    tol = np.finfo_eps(approximation.dtype)
    assert np.allclose(approximation @ approximation.T, cov, atol=tol, rtol=tol)


@testing.parametrize_with_cases("cholesky", cases=".", prefix="case_cholesky")
def test_output_the_right_shapes(cholesky: Callable, n=4, rank=4):
    key = prng.prng_key(1)

    cov_eig = 0.1 + prng.uniform(key, shape=(n,))
    cov = test_util.symmetric_matrix_from_eigenvalues(cov_eig)

    approximation, _info = cholesky(lambda i, j: cov[i, j], nrows=n, rank=rank)()
    assert approximation.shape == (n, rank)


def test_full_rank_nopivot_matches_cholesky(n=10):
    key = prng.prng_key(2)
    cov_eig = 0.01 + prng.uniform(key, shape=(n,), dtype=float)
    cov = test_util.symmetric_matrix_from_eigenvalues(cov_eig)
    reference = linalg.cholesky(cov)

    # Sanity check: pivoting should definitely not satisfy this:
    cholesky_p = low_rank.cholesky_partial_pivot(
        lambda i, j: cov[i, j], nrows=n, rank=n
    )
    received, info = cholesky_p()
    assert not np.allclose(received, reference)

    # But without pivoting, we should get there!
    cholesky = low_rank.cholesky_partial(lambda i, j: cov[i, j], nrows=n, rank=n)
    received, info = cholesky()
    assert np.allclose(received, reference, atol=1e-6)


def test_pivoting_improves_the_estimate(n=10, rank=5):
    key = prng.prng_key(1)

    cov_eig = 0.1 + prng.uniform(key, shape=(n,))
    cov = test_util.symmetric_matrix_from_eigenvalues(cov_eig)

    def element(i, j):
        return cov[i, j]

    nopivot, _info = low_rank.cholesky_partial(element, nrows=n, rank=rank)()
    pivot, _info = low_rank.cholesky_partial_pivot(element, nrows=n, rank=rank)()

    error_nopivot = linalg.matrix_norm(cov - nopivot @ nopivot.T, which="fro")
    error_pivot = linalg.matrix_norm(cov - pivot @ pivot.T, which="fro")
    assert error_pivot < error_nopivot
